/*
MagicOverlayRenderer class
*/
#include "MagicOverlayRenderer.h"
#include "colorspaces.h"
#include "MediaView.h"
#include "rgb_to_ycbcr.h"
#include <alloc.h>
#include <stdio.h>

#ifndef uchar
typedef unsigned char uchar;
#endif

MagicOverlayRenderer::MagicOverlayRenderer()
{
	drawer = -1;
	decode_sem = 0;
	draw_sem = 0;
	droppedframes=0;
	bitmap[0] = NULL;
	bitmap[1] = NULL;
	bitmapBounds = NULL;
	rgb_sz = 0;
	drawbm_sz = 0;
	view = NULL;
	cs_index = -1;
	cspace_used = B_NO_COLOR_SPACE;
}

MagicOverlayRenderer::~MagicOverlayRenderer()
{
	if(view)
		Dispose();
}

const char *MagicOverlayRenderer::ColorSpaceName() const
{
	if(cs_index == -1)
		return "none";
	return colspace[cs_index].name;
}

color_space MagicOverlayRenderer::ColorSpace()
{
	if(cs_index == -1)
		return B_NO_COLOR_SPACE;
	return colspace[cs_index].colspace;
}

void *MagicOverlayRenderer::Bits()
{
	if(!bitmap[BM_DECODE])
		return NULL;
	return bitmap[BM_DECODE]->Bits();
}

status_t MagicOverlayRenderer::LockBits()
{
	return bitmap[BM_DECODE]->LockBits();
}

status_t MagicOverlayRenderer::UnlockBits()
{
	bitmap[BM_DECODE]->UnlockBits();
	return B_OK;
}
	
	// render the frame in buff, the next buffer to write to is returned (can be the same)
	// should block until the buffer returned is ready to be fulled
void *MagicOverlayRenderer::Draw(void *buff)
{
	status_t err = B_OK;
////	if((err=acquire_sem_etc(decode_sem, 1, B_RELATIVE_TIMEOUT, 500000)) == B_OK) {
	if((err=acquire_sem(decode_sem)) == B_OK) {
////		if(!bitmap || !view)
////			break;
/*
	bitmap[BM_DECODE]->LockBits();
	bitmap[BM_DRAW]->LockBits();
	memcpy(bitmap[BM_DRAW]->Bits(), bitmap[BM_DECODE]->Bits(), bm_sz); // we only copy the Bits(), not the BBitmaps :)))) (I should go to bed now)
	bitmap[BM_DRAW]->UnlockBits();
	bitmap[BM_DECODE]->UnlockBits();
*/		
		release_sem(draw_sem);
//	} else if(err == B_TIMEOUT) {
//		droppedframes++;
	}


	return bitmap[BM_DECODE]->Bits();
}
	
status_t MagicOverlayRenderer::Install(MediaView *view, media_format *format)
{
	bool found = false;
	int i=0;

	puts(">install");

	droppedframes=0;

	if(!format)
		return B_ERROR;
	
	if(format->type != B_MEDIA_RAW_VIDEO/* && format->type != B_MEDIA_ENCODED_VIDEO*/)
		return B_ERROR;

	DUMP_FORMAT(format);
	
	if(format->u.raw_video.display.format != B_RGB32)
		return B_ERROR; // only RGB32 -> YCbCr (for now)
	
	for(i=0; !found && i<SUPP_OVL; i++)
		if(colspace[i].colspace == format->u.raw_video.display.format)
			found = true;
	
	if(!found)
		return B_ERROR;

	i--;
	cs_index = i;
	
	found = false;
	
/*	bitmapBounds = new BRect (0.0, 
						   0.0, 
						   format->u.encoded_video.output.display.line_width - 1.0,
						   format->u.encoded_video.output.display.line_count - 1.0);
*/
	bitmapBounds = new BRect (0.0, 
						   0.0, 
						   format->u.raw_video.display.line_width - 1.0,
						   format->u.raw_video.display.line_count - 1.0);

	bitmap[BM_DECODE] = new BBitmap(*bitmapBounds, B_BITMAP_ACCEPTS_VIEWS, colspace[cs_index].colspace);
	if(!bitmap[BM_DECODE] || bitmap[BM_DECODE]->InitCheck() != B_OK) {
		puts("not supported. (1)");
		if(bitmap[BM_DECODE]) delete bitmap[BM_DECODE];
		bitmap[BM_DECODE] = NULL;
		puts("not supported. (1)");
		return B_ERROR;
	}

	// first try B_RGB16
	cspace_used = B_RGB16;

#ifdef HACK_BGA // ;-)
	bitmap[BM_DRAW] = new BBitmap(*bitmapBounds,B_BITMAP_WILL_OVERLAY, B_RGB16);
#else
	bitmap[BM_DRAW] = new BBitmap(*bitmapBounds,B_BITMAP_WILL_OVERLAY|B_BITMAP_RESERVE_OVERLAY_CHANNEL, B_RGB16);
#endif
	if(!bitmap[BM_DRAW] || bitmap[BM_DRAW]->InitCheck() != B_OK) {

		// it failed... try B_YCbCr422
		cspace_used = B_YCbCr422;

		puts("B_RGB16 not supported. (2)");
//		if(bitmap[BM_DECODE]) delete bitmap[BM_DECODE];
//		bitmap[BM_DECODE] = NULL;
		if(bitmap[BM_DRAW]) delete bitmap[BM_DRAW];
		bitmap[BM_DRAW] = NULL;
		
	
#ifdef HACK_BGA // ;-)
		bitmap[BM_DRAW] = new BBitmap(*bitmapBounds,B_BITMAP_WILL_OVERLAY, B_YCbCr422);
#else
		bitmap[BM_DRAW] = new BBitmap(*bitmapBounds,B_BITMAP_WILL_OVERLAY|B_BITMAP_RESERVE_OVERLAY_CHANNEL, B_YCbCr422);
#endif
		if(!bitmap[BM_DRAW] || bitmap[BM_DRAW]->InitCheck() != B_OK) {
			puts("B_YCbCr422 not supported. (2)");
			if(bitmap[BM_DECODE]) delete bitmap[BM_DECODE];
			bitmap[BM_DECODE] = NULL;
			if(bitmap[BM_DRAW]) delete bitmap[BM_DRAW];
			bitmap[BM_DRAW] = NULL;
			return B_ERROR;
		} else
			puts("using B_YCbCr422 overlay");
	} else
		puts("using B_RGB16 overlay");

	rgb_sz = bitmap[BM_DECODE]->BitsLength();
	drawbm_sz = bitmap[BM_DRAW]->BitsLength();

	puts("supported.");

	this->view = view;

	draw_sem = create_sem(0, "dblbuff_dec");
	decode_sem = create_sem(0, "dblbuff_draw");

	// we have an YCbCr overlay, we need some tables to do the conversion
	if(cspace_used == B_YCbCr422) {
		lt_r_y = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_g_y = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_b_y = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_r_cb = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_g_cb = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_b_cb = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_r_cr = lt_b_cb; // same as lt_b_cb
		lt_g_cr = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
		lt_b_cr = (int32 *)malloc(TABLE_SIZE * SIZEOF(INT32));
	}

	
	thread_func convert_func = (cspace_used == B_RGB16)?(MagicOverlayRenderer::DrawerThreadRGB16):(MagicOverlayRenderer::DrawerThreadYCbCr422);

	drawer = spawn_thread(convert_func, "drawer", PRIO_DRAWER, this);
	if(drawer == 0) {
/*		if(bitmap[0]) delete bitmap[0];
		bitmap[0] = NULL;
		if(bitmap[1]) delete bitmap[1];
		bitmap[1] = NULL;
		delete_sem(draw_sem);
		delete_sem(decode_sem);
*/
		Dispose(); // it's sometimes so simple... :^)
		return B_ERROR;
	}
	
	resume_thread(drawer);

	if(view->Window()->LockWithTimeout(50000) == B_OK) {
		view->Picture()->Invalidate(view->Picture()->Bounds());
		view->Invalidate(view->Bounds());
		view->Flush();
		view->Window()->Unlock();
	}
	puts("<install");
	return B_OK;
}

status_t MagicOverlayRenderer::Use()
{
	int i;

	puts(">use");

	if(!view)
		return B_ERROR;
	if(cspace_used == B_YCbCr422) {
		for (i = 0; i <= MAXSAMPLE; i++) {

#ifdef NEVRAX_EQS
			lt_r_y[i] = FIX(0.29900) * i;
			lt_g_y[i] = FIX(0.58700) * i;
			lt_b_y[i] = FIX(0.11400) * i     + ONE_HALF;
			lt_r_cb[i] = (-FIX(0.16874)) * i;
			lt_g_cb[i] = (-FIX(0.33126)) * i;
			/* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
			 * This ensures that the maximum output will round to MAXJSAMPLE
			 * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
			 */
			lt_b_cb[i] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
			/*  B=>Cb and R=>Cr tables are the same
			    rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
			 */
			lt_g_cr[i] = (-FIX(0.41869)) * i;
			lt_b_cr[i] = (-FIX(0.08131)) * i;
#else
			lt_r_y[i] = FIX(0.25700) * i;
			lt_g_y[i] = FIX(0.50400) * i;
			lt_b_y[i] = FIX(0.09800) * i     + ONE_HALF + Y_OFFSET;
			lt_r_cb[i] = (-FIX(0.14800)) * i;
			lt_g_cb[i] = (-FIX(0.29100)) * i;
			/* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
			 * This ensures that the maximum output will round to MAXJSAMPLE
			 * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
			 */
			lt_b_cb[i] = FIX(0.43900) * i    + CBCR_OFFSET + ONE_HALF-1;
			/*  B=>Cb and R=>Cr tables are the same
			    rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
			 */
			lt_g_cr[i] = (-FIX(0.36800)) * i;
			lt_b_cr[i] = (-FIX(0.07100)) * i;
#endif
		}
	}
	
	overlay_restrictions r;
	bitmap[BM_DRAW]->GetOverlayRestrictions(&r);
	printf("Using overlay: %s (%04X)\n", colspace[cs_index].name, colspace[cs_index]);

	printf("Overlay limits:\n");
	printf("  Src horiz alignment  : %08x\n",r.source.horizontal_alignment);
	printf("  Src vert alignment   : %08x\n",r.source.vertical_alignment);
	printf("  Src width alignment  : %08x\n",r.source.width_alignment);
	printf("  Src height alignment : %08x\n",r.source.height_alignment);
	printf("  Src min/max          : (%d,%d)/(%d,%d)\n",r.source.min_width,r.source.min_height,
														r.source.max_width,r.source.max_height);
	printf("  Dst horiz alignment  : %08x\n",r.destination.horizontal_alignment);
	printf("  Dst vert alignment   : %08x\n",r.destination.vertical_alignment);
	printf("  Dst width alignment  : %08x\n",r.destination.width_alignment);
	printf("  Dst height alignment : %08x\n",r.destination.height_alignment);
	printf("  Dst min/max          : (%d,%d)/(%d,%d)\n",r.destination.min_width,r.destination.min_height,
														r.destination.max_width,r.destination.max_height);
	printf("  Min/max scaling      : (%f,%f)/(%f,%f)\n",r.min_width_scale,r.min_height_scale,
														r.max_width_scale,r.max_height_scale);
	rgb_color key;
	if(view && view->Window()!=NULL && view->Window()->LockWithTimeout(50000) == B_OK) {
		view->Picture()->SetViewOverlay(bitmap[BM_DRAW],*bitmapBounds,view->VideoBounds(),&key,B_FOLLOW_ALL,
			B_OVERLAY_FILTER_HORIZONTAL|B_OVERLAY_FILTER_VERTICAL);
		view->Picture()->SetViewColor(key);
//#ifdef USE_SUBVIEW
		view->Picture()->Invalidate(view->Picture()->Bounds());
//#endif
		view->Invalidate(view->Bounds());
		view->Flush();
		view->Window()->Unlock();
	}
	release_sem(decode_sem);
	puts("<use");
	return B_OK;
}

status_t MagicOverlayRenderer::Dispose()
{
	puts(">dispose");
	if(!bitmap)
		return B_ERROR;
	status_t err;
	
	if(!bitmap[0] || !bitmap[1])
		return B_ERROR;
	if(decode_sem)
		delete_sem(decode_sem);
	decode_sem=0;
	if(draw_sem)
		delete_sem(draw_sem);
	draw_sem=0;
	if(drawer != -1)
		wait_for_thread(drawer, &err);
	drawer = -1;
	if(view && view->Window()!=NULL && view->Window()->Lock()) {
		view->Picture()->ClearViewOverlay();
		view->Picture()->SetViewColor(B_TRANSPARENT_32_BIT);
/*		if(view->Window()->Lock()) {
			view->Invalidate(view->Bounds());
			view->Flush();
			view->Window()->Unlock();
		}*/
		view->Window()->Unlock();
	}
//	bitmap->UnlockBits();
	if(bitmap[0])
		delete bitmap[0];
	bitmap[0] = NULL;
	if(bitmap[1])
		delete bitmap[1];
	bitmap[1] = NULL;

	if(lt_r_y)
		free(lt_r_y);
	lt_r_y = NULL;
	if(lt_g_y)
		free(lt_g_y);
	lt_g_y = NULL;
	if(lt_b_y)
		free(lt_b_y);
	lt_b_y = NULL;
	if(lt_r_cb)
		free(lt_r_cb);
	lt_r_cb = NULL;
	if(lt_g_cb)
		free(lt_g_cb);
	lt_g_cb = NULL;
	if(lt_b_cb)
		free(lt_b_cb);
	lt_b_cb = NULL;
	if(lt_g_cr)
		free(lt_g_cr);
	lt_g_cr = NULL;
	if(lt_b_cr)
		free(lt_b_cr);
	lt_b_cr = NULL;

	puts("<dispose");
	return B_OK;
}

color_space MagicOverlayRenderer::NextColorSpace(int *cookie) // cookie = 0 first
{
	if(*cookie >= SUPP_OVL)
		return B_NO_COLOR_SPACE;
	return colspace[(*cookie)++].colspace;
}

BRect MagicOverlayRenderer::Bounds()
{
	if(!bitmap[0])
		return BRect();
	return bitmap[0]->Bounds();
}

int32 MagicOverlayRenderer::BytesPerRow()
{
	if(!bitmap[0])
		return 0;
	return bitmap[0]->BytesPerRow();
}


int32 MagicOverlayRenderer::DrawerThreadRGB16(void *arg)
{
	MagicOverlayRenderer *renderer;
	MediaView *view;
	
	uchar *rgbplane;
	int32 *overlay;
	int32 *ovl_end;
	
	renderer = (MagicOverlayRenderer *)arg;
	view = renderer->view;

	
//	register long cnt;
	register int r, g, b;
	register int r1, g1, b1;
	register uchar y0, y1, cb, cr;
	
	register int32 ov_long;
	register int32 rgb_long;
	
	printf("RGB_SZ=%ld drawbm_sz=%ld\n", renderer->rgb_sz, renderer->drawbm_sz);
	
	while(acquire_sem(renderer->draw_sem) == B_OK) {
		if(!renderer->bitmap[BM_DRAW] || !view)
			break;
//		puts("MOR - got lock");
			
		renderer->bitmap[BM_DECODE]->LockBits();
		renderer->bitmap[BM_DRAW]->LockBits();
		
		rgbplane = (uchar *)(renderer->bitmap[BM_DECODE])->Bits(); // stupid since it doesn't move...
		overlay = (int32 *)(renderer->bitmap[BM_DRAW])->Bits();
		
		ovl_end = (int32 *)(((char *)overlay)+(renderer->drawbm_sz));

		// DEBGU
//		ovl_end = (int32 *)(((char *)overlay)+(renderer->drawbm_sz/2));


//		puts("begin rgb2ycc");


/*
		dbg1 = 0;
		dbg2 = 0;
		dbg3 = 0;
*/

		for(; overlay<ovl_end; overlay++, rgbplane+=4) {

			b = *(rgbplane);
			g = *(rgbplane+1);
			r = *(rgbplane+2);
			
			
			// from GraphicsDefs.h
			// B_RGB16 =                       0x0005,
			// MSB                      LSB
			// G[2:0],B[4:0]  R[4:0],G[5:3]
	//		ov_long = (((((long)(g>>2)) & 0x07)<<13) | ((((long)(b>>3)) & 0x01F)<<8) | ((((long)(r>>3)) & 0x01F)<<3) | (((long)(g>>5)) & 0x07));
			
	//		ov_long = ov_long<<16;
			// do it for the next pixel too, so we have an int32 to put in the overlay (faster)
			rgbplane+=4;
			b1 = *(rgbplane);
			g1 = *(rgbplane+1);
			r1 = *(rgbplane+2);


			// original code
	//		ov_long = (ov_long & 0x0FFFF0000) | (((((long)(g1>>2)) & 0x07)<<13) | ((((long)(b1>>3)) & 0x01F)<<8) | ((((long)(r1>>3)) & 0x01F)<<3) | (((long)(g1>>5)) & 0x07));
			
			// thanks to BGA
			ov_long = ((r << 8) & (0x1f << 11)) + ((g << 3) & (0x3f << 5)) + ((b >> 3) & 0x1f) + ((r1 << 24) & (0x1f << 27)) + ((g1 << 19) & (0x3f << 21)) + ((b1 << 13) & (0x1f << 16));
			
			*(overlay) = ov_long;
		}
		
		renderer->bitmap[BM_DRAW]->UnlockBits();
		renderer->bitmap[BM_DECODE]->UnlockBits();
		release_sem(renderer->decode_sem);
	}
	return B_OK;
}

int32 MagicOverlayRenderer::DrawerThreadYCbCr422(void *arg)
{
	MagicOverlayRenderer *renderer;
	MediaView *view;
	
	int dbg1, dbg2, dbg3;
	
	uchar *rgbplane;
	int32 *overlay;
	int32 *ovl_end;
	
	renderer = (MagicOverlayRenderer *)arg;
	view = renderer->view;

	int32		*lt_r_y = renderer->lt_r_y;
	int32		*lt_g_y = renderer->lt_g_y;
	int32		*lt_b_y = renderer->lt_b_y;
	int32		*lt_r_cb = renderer->lt_r_cb;
	int32		*lt_g_cb = renderer->lt_g_cb;
	int32		*lt_b_cb = renderer->lt_b_cb;
	int32		*lt_r_cr = lt_b_cb; // same as lt_b_cb
	int32		*lt_g_cr = renderer->lt_g_cr;
	int32		*lt_b_cr = renderer->lt_b_cr;

	
//	register long cnt;
	register int r, g, b;
	register uchar y0, y1, cb, cr;
	
	register int32 ov_long;
	register int32 rgb_long;
	
	printf("RGB_SZ=%ld drawbm_sz=%ld\n", renderer->rgb_sz, renderer->drawbm_sz);
	
	while(acquire_sem(renderer->draw_sem) == B_OK) {
		if(!renderer->bitmap[BM_DRAW] || !view)
			break;
//		puts("MOR - got lock");
			
		renderer->bitmap[BM_DECODE]->LockBits();
		renderer->bitmap[BM_DRAW]->LockBits();
		
		rgbplane = (uchar *)(renderer->bitmap[BM_DECODE])->Bits(); // stupid since it doesn't move...
		overlay = (int32 *)(renderer->bitmap[BM_DRAW])->Bits();
		
		ovl_end = (int32 *)(((char *)overlay)+(renderer->drawbm_sz));

		// DEBGU
//		ovl_end = (int32 *)(((char *)overlay)+(renderer->drawbm_sz/2));


//		puts("begin rgb2ycc");


/*
		dbg1 = 0;
		dbg2 = 0;
		dbg3 = 0;
*/

		for(; overlay<ovl_end; overlay++, rgbplane+=4) {

			b = *(rgbplane);
			g = *(rgbplane+1);
			r = *(rgbplane+2);

/*			rgb_long = *((int32 *)rgbplane);
			r = rgb_long
*/			
			y0 = (uchar)((lt_r_y[r] + lt_g_y[g] + lt_b_y[b]) >> SCALEBITS);
//			y0>>1;
//			ov_long = y0;
//			ov_long << 8;
			
#ifndef RGBYCC_NO_MEAN
			cb = (uchar)((lt_r_cb[r] + lt_g_cb[g] + lt_b_cb[b]) >> (SCALEBITS+1));
			cr = (uchar)((lt_r_cr[r] + lt_g_cr[g] + lt_b_cr[b]) >> (SCALEBITS+1));
#else
			cb = (uchar)((lt_r_cb[r] + lt_g_cb[g] + lt_b_cb[b]) >> (SCALEBITS));
			cr = (uchar)((lt_r_cr[r] + lt_g_cr[g] + lt_b_cr[b]) >> (SCALEBITS));
#endif
			rgbplane+=4;
			b = *(rgbplane);
			g = *(rgbplane+1);
			r = *(rgbplane+2);

#ifndef RGBYCC_NO_MEAN
			cb += (uchar)((lt_r_cb[r] + lt_g_cb[g] + lt_b_cb[b]) >> (SCALEBITS+1));
			cr += (uchar)((lt_r_cr[r] + lt_g_cr[g] + lt_b_cr[b]) >> (SCALEBITS+1));
#endif
			y1 = (uchar)((lt_r_y[r] + lt_g_y[g] + lt_b_y[b]) >> SCALEBITS);

//			y1>>1;
//			cb>>1;
//			cr>>1;
			
//			cb = 128;
//			cr = 128;
/*
			dbg1++;
			if(dbg1==256) {
				dbg1=0;
				dbg2++;
				if(dbg2==256) {
					dbg2=0;
					dbg3++;
					dbg3 = dbg3 % 256;
				}
			}
*/
/*			y0 = dbg1;
			y1 = dbg1;
			cb = dbg1;
			cr = dbg1;
			y0 = 235;
			y1 = 128;
*/			
/*
			ov_long = y0;
			ov_long << 8;
			ov_long |= cb;
			ov_long << 8;
			ov_long |= y1;
			ov_long << 8;
			ov_long |= cr;
*/
			// LITTLE ENDIAN !!!
//			ov_long = 0L;
/*
// doesn't work ??
//			ov_long |= (unsigned long)cr;
			ov_long << 8;
//			ov_long |= (unsigned long)y1;
			ov_long << 8;
			ov_long |= (unsigned long)cb;
			ov_long << 8;
//			ov_long |= (unsigned long)y0;
*/

			ov_long = ((y0&0x0FF) | ((cb<<8)&0x0FF00) | ((y1<<16)&0x0FF0000) | ((cr<<24)&0x0FF000000));
			
			//ov_long << 8;
			
			
			//ov_long
			// seems writing longs is faster than byte/byte
			// (dixit stampTV source)
			*(overlay) = ov_long;
		}
		
		renderer->bitmap[BM_DRAW]->UnlockBits();
		renderer->bitmap[BM_DECODE]->UnlockBits();
//		puts("done rgb2ycc");
		release_sem(renderer->decode_sem);
	}
	return B_OK;
}

